/* hardware.h */

#ifdef __APPLE__

#include <IOKit/IOKitLib.h>
#include <IOKit/IOCFPlugIn.h>
#include <IOKit/scsi-commands/SCSITaskLib.h>
#include <CoreFoundation/CFString.h>
#include <CoreFoundation/CFDictionary.h>
#include <CoreFoundation/CFData.h>

typedef struct {
	unsigned char reserved;
	unsigned char adr_control;
	unsigned char track;
	unsigned char reserved2;
	unsigned char reserved3;
	unsigned char min, sec, frame;
} MMCTOCEntry;

typedef struct {
	unsigned char len_high, len_low;
	unsigned char first_track, last_track;
	MMCTOCEntry entries[100];
} MMCTOC;

typedef struct {
	unsigned char session;
	unsigned char adr_control;
	unsigned char tno, point;
	unsigned char min, sec, frame;
	unsigned char zero;
	unsigned char pmin, psec, pframe;
} MMCTOCFMT2Entry;

typedef struct {
	unsigned char len_high, len_low;
	unsigned char first_track, last_track;
	MMCTOCFMT2Entry entries[100];
} MMCTOCFMT2;

extern MMCTOC mmc_toc;
extern io_service_t obj;
extern MMCDeviceInterface **mmc;
extern IOCFPlugInInterface **plugin;

int osx_open_cdrom(const char *name);
void osx_close_cdrom(void);

// Linux compatibility stuff

struct cdrom_tochdr {
	unsigned char cdth_trk0, cdth_trk1;
};

struct cdrom_msf0 {
	unsigned char minute, second, frame;
};

union cdrom_addr {
	struct cdrom_msf0 msf;
	int lba;
};

struct cdrom_tocentry {
	unsigned char cdte_track;
	unsigned char cdte_adr:4, cdte_ctrl:4;
	unsigned char cdte_format;
	union cdrom_addr cdte_addr;
	union cdrom_addr cdte_addr_end;
	unsigned char cdte_datamode;
};

struct cdrom_subchnl {
	unsigned char cdsc_format;
	unsigned char cdsc_audiostatus;
	unsigned char cdsc_adr:4, cdsc_ctrl:4;
	unsigned char cdsc_trk, cdsc_ind;
	union cdrom_addr cdsc_absaddr, cdsc_reladdr;
};

#define CDROM_MSF 2
#define CDROM_DATA_TRACK 4

#define CDROM_AUDIO_INVALID   0x00
#define CDROM_AUDIO_PLAY      0x11
#define CDROM_AUDIO_PAUSED    0x12
#define CDROM_AUDIO_COMPLETED 0x13
#define CDROM_AUDIO_ERROR     0x14
#define CDROM_AUDIO_NO_STATUS 0x15

// Returns 1 on success, 0 on no-disc or other error
int osx_read_toc(MMCDeviceInterface **mmc, struct cdrom_tocentry *entries,
                 int nentries, struct cdrom_tochdr *tochdr);

#endif /* __APPLE__ */

/* state of the CD-ROM hardware */
typedef struct {
    int    iStatus;                        /* see CD_STATUS* below */
    int    iData;
    int    iAudio;
    struct cdrom_tochdr tochdr;            /* from CDROMREADTOCHDR */
    struct cdrom_tocentry tocentries[100]; /* from CDROMREADTOCENTRY */
    struct cdrom_subchnl subchnl;          /* from CDROMSUBCHNL */
} cdhw_t;

/* procedures */
cdhw_t *read_hw(char *progname, int cdfile, char *cdname);
int     checkmount(char *progname, char *pszName);
int     cd_status(char *progname, int cd_fd, char *cdname, cdhw_t *pHw);
void    free_hw_buf(cdhw_t *pHw);
char   *cd_status_text (int iStatus);

/* additional error codes, see linux/cdrom.h */
#define NO_CDROM         200
#define CDROM_MOUNTED    201
#define CDROM_DATA_DISC  202

/* return codes from cd_status */
#define CD_STATUS_ERROR    -1
#define CD_STATUS_NO_CD     0
#define CD_STATUS_CD_IN     1
#define CD_STATUS_INSERTED  2
#define CD_STATUS_REMOVED   3
#define CD_STATUS_MOUNTED   4
#define CD_STATUS_DATA_DISC 5


